/*
   statement.c -- PIKT alarm script statement routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/statement.c,v 1.52 2001/03/24 05:21:49 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2001 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"

#include "piktd.h"

/* perform various actions for current program statement; in simple
   cases (such as REDO), just return with a value indicating where next
   (what statement) to go in program (quite often to immediate next
   program statement (as with ENDIF, for example); in complex cases,
   current statement has expressions to be evaluated; in the case of
   setting an array variable, an expression on left-hand side of "="
   needs to be evaluated; in most other cases (and also in array setting
   case), a right-hand side expression must be evaluated; after expression
   evaluation, perform various actions appropriate to current statement,
   using evaluated expressions; in all cases, return a value indicating
   where next (what statement) to go in program flow for, example TRUE
   (for tnext statement), FALSE (for fnext statement), AGAIN, LEAVE, REDO,
   NEXT, LAST, QUIT, etc.
 */
int
dostatement (int type)
{
	static char fnnam[] = "dostatement";
	char *cmd;
	pid_t pid;
	vl *vlp;
	register int k;

	switch (type) {
		case AGAIN:	/* apply current rule again */
			return AGAIN;
			break;
		case LEAVE:	/* leave current rule and go on to next */
			return LEAVE;
			break;
		case REDO:	/* redo current line */
			return REDO;
			break;
		case NEXT:	/* go on to next line */
			return NEXT;
			break;
		case LAST:	/* go to end */
			return LAST;
			break;
		case QUIT:	/* go on to next alarm */
			return QUIT;
			break;
		case ELSE:
		case ENDIF:
		case ENDWHILE:
		case WHILEAUX:
		case REPEAT:
		case REPEATAUX:
		case ENDFOR:
		case FORAUX:
		case SWITCH:
		case DEFAULT:
		case ENDSWITCH:
			return TRUE;
			break;
		case BREAK:
		case CONTINUE:
		case BREAKSWITCH:
			return FALSE;
			break;
		case SET:
		case UNSET:
			if (stmcurr->lhsexpr) {
				tokcurr = stmcurr->lhsexpr;
				exprres = NIL;
				exprval = NIL;
				exprparse();
				if (! exprstr) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, null exprstr", fnnam, ALMNAM);
					STATUS = INACTIVE;
					return QUIT;
				}
				if (exprval != ERRARNDX) {
					for (k=0; k<=MAXARYDIM; k++)
						stmcurr->arndx[k] = arndx[k];
					FREE(stmcurr->arkey);
					stmcurr->arkey = mystrdup(exprstr, fnnam);
				} else {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, bad array index: %f", fnnam, ALMNAM, exprval);
					for (k=0; k<=MAXARYDIM; k++)
						stmcurr->arndx[k] = ERRARNDX;
					FREEVOID(stmcurr->arkey);
					return TRUE;
				}
			}
			/* no break here! */
		default:	/* all other cases have rhsexprs */
			tokcurr = stmcurr->rhsexpr;
			break;
	}
	if (exprres == TRUE) exprres = NIL;
	exprval = NIL;
	exprparse();
	if (! exprstr) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, null exprstr", fnnam, ALMNAM);
		STATUS = INACTIVE;
		return QUIT;
	}
	requote(exprstr);
	switch (type) {
		case DO:
			return TRUE;
			break;
		case IF:
		case ELIF:
		case WHILE:
		case UNTIL:
		case FOR:
		case CASE:
			if (exprres == TRUE)
				return TRUE;
			else
				return FALSE;
			break;
		case SKIP:	/* go ahead n lines */
			skip = (int) exprval;
			return SKIP;
			break;
		case PAUSE:
			if ((int)exprval >= 0)
				sleep((unsigned int)exprval);
			else
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, illegal (negative or zero) pause argument: %f", fnnam, ALMNAM, exprval);
			return TRUE;
			break;
		case DIE:	/* exit program */
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, bad die line: %s", fnnam, ALMNAM, exprstr);
			}
			if (isautosyslog)
				logreturn(SYSLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, %s", fnnam, ALMNAM, exprstr);
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, %s", fnnam, ALMNAM, exprstr);
			break;
		case OUTPUT:
		case MAIL:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, line not output, error in line: %s", fnnam, ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
				return TRUE;
			} else {
				if (! interactive) {
					if (newmailalert) {
						althdr(&mailmsg, type);
						newmailalert = FALSE;
					}
					if (newmailalarm) {
						alarmhdr(&mailmsg, alarmcurr, type);
						newmailalarm = FALSE;
					}
				}
				altlin(&mailmsg, exprstr, type);
				if (isautosyslog && (type != OUTPUT))
					logreturn(SYSLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
				return TRUE;
			}
			break;
		case PRINT:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, line not output, error in line: %s", fnnam, ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
				return TRUE;
			} else {
				if (! interactive) {
					if (newprintalert) {
						althdr(&printmsg, type);
						newprintalert = FALSE;
					}
					if (newprintalarm) {
						alarmhdr(&printmsg, alarmcurr, type);
						newprintalarm = FALSE;
					}
				}
				altlin(&printmsg, exprstr, type);
				if (isautosyslog)
					logreturn(SYSLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
				return TRUE;
			}
			break;
		case LOCLOG:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, bad log line: %s", fnnam, ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			if (strcmp(stmcurr->logfile, "/dev/null")) {
				logreturn(LOCLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
				/* but what if write failure to LOCLOG? */
				if (isautosyslog)
					logreturn(SYSLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
			}
			return TRUE;
			break;
		case PIKTLOG:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, bad log line: %s", fnnam, ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			logreturn(PIKTLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
			if (isautosyslog)
				logreturn(SYSLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
			return TRUE;
			break;
		case SYSLOG:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, bad log line: %s", fnnam, ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			logreturn(SYSLOG, LEVEL, __FILE__, __LINE__, FALSE, "%s", exprstr);
			return TRUE;
			break;
		case SET:
			if (INPUTTMPFIL && (stmcurr->vrndx == OUTLINNOW)) {
				REWRITEINFIL = TRUE;
				outlinnow = mystrdup(exprstr, fnnam);
				/* FREEVOID(stmcurr->arkey); */
				return TRUE;
			}
			if (VALTYP(stmcurr->vrndx) & FIXEDVAL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, attempt to set data variable %s", fnnam, ALMNAM, VARNAM(stmcurr->vrndx));
				STATUS = INACTIVE;
				FREEVOID(stmcurr->arkey);
				return QUIT;
			}
			if (VALTYP(stmcurr->vrndx) & BEGINVAL) {
				if ((vlp = lookupvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
					FREE(vlp->str);
				} else {
					if (! (vlp = regvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, regvl() failure for var %s", fnnam, ALMNAM, VARNAM(stmcurr->vrndx));
						STATUS = INACTIVE;
						FREEVOID(stmcurr->arkey);
						return QUIT;
					}
				}
				vlp->str = mystrdup(exprstr, fnnam);
				vlp->num = exprval;
				/* should be unnecessary:
				vlp->fil = (FILE *)NULL;
				 */
			}
			if (VALTYP(stmcurr->vrndx) & (BEGINVAL|FREEVAL|ENDVAL)) {
				if ((vlp = lookupvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
					FREE(vlp->str);
				} else {
					if (! (vlp = regvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, regvl() failure for var %s", fnnam, ALMNAM, VARNAM(stmcurr->vrndx));
						STATUS = INACTIVE;
						FREEVOID(stmcurr->arkey);
						return QUIT;
					}
				}
				vlp->str = mystrdup(exprstr, fnnam);
				vlp->num = exprval;
				/* should be unnecessary:
				vlp->fil = (FILE *)NULL;
				 */
				MAPKEYSVRNDXARY(stmcurr->vrndx) = NIL;
			}
			if (isdebug) {
				vlp = lookupvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
				if (vlp && vlp->str) {
					printf("DEBUG: %s (str) set to: %s\n", VARNAM(stmcurr->vrndx), vlp->str);
					fflush(stdout);
				}
				if (vlp && (vlp->num != NIL)) {
					printf("DEBUG: %s (num) set to: %f\n", VARNAM(stmcurr->vrndx), vlp->num);
					fflush(stdout);
				}
			}
			FREEVOID(stmcurr->arkey);
			return TRUE;
			break;
		case UNSET:
			/*
			if (VALTYP(stmcurr->vrndx) & FIXEDVAL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, attempt to unset data variable %s", fnnam, ALMNAM, VARNAM(stmcurr->vrndx));
				STATUS = INACTIVE;
				FREEVOID(stmcurr->arkey);
				return QUIT;
			}
			*/
			if (VALTYP(stmcurr->vrndx) & BEGINVAL) {
				delvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
			}
			if (VALTYP(stmcurr->vrndx) & (BEGINVAL|FREEVAL|ENDVAL)) {
				delvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
			}
			if (isdebug) {
				printf("DEBUG: %s[%s] unset\n", VARNAM(stmcurr->vrndx), stmcurr->arkey ? stmcurr->arkey : "");
				fflush(stdout);
			}
			delarkey(stmcurr->vrndx, &(ARKEYS(stmcurr->vrndx)), stmcurr->arkey, FALSE);
			delarkey(stmcurr->vrndx, &(SARKEYS(stmcurr->vrndx)), stmcurr->arkey, TRUE);
			FREEVOID(stmcurr->arkey);
			return TRUE;
			break;
		case EXEC:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, command not exec'ed, error in command: %s", fnnam, ALMNAM, exprstr);
				STATUS = INACTIVE;
				return QUIT;
			} else {
				if (! *exprstr) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, null statement string", fnnam, ALMNAM);
					return TRUE;
				}
				switch (pid = fork()) {
					case -1:	/* error */
						logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in %s(), %s, fork() error", fnnam, ALMNAM);
						break;
					case 0:		/* child */
						cmd = (char *) mymalloc(strlen(exprstr)+strlen(piktlog)+6, fnnam);
						sprintf(cmd, "%s 2>> %s", exprstr, piktlog);
						if (isautosyslog)
							logreturn(SYSLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "in %s(), %s, exec \"%s\"", fnnam, ALMNAM, cmd);
						logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "in %s(), %s, exec \"%s\"", fnnam, ALMNAM, cmd);
						execl(SHELL, "sh", "-c", cmd, (char *) NULL);
						free(cmd);
						_exit(127);
						break;
					default:	/* parent */
						break;
				}
				return TRUE;
			}
			break;
		case EXECWAIT:
			if (stmcurr->status == INACTIVE) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, command not exec'ed, error in command: %s", fnnam, ALMNAM, exprstr);
				STATUS = INACTIVE;
				return QUIT;
			} else {
				if (! *exprstr) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, null statement string", fnnam, ALMNAM);
					return TRUE;
				}
				/* oh boy, more lock files!  i am reluctant
				   to do this, since lock file management
				   can get messy, but i see no other way;
				   we don't check for a lock file here,
				   because we don't want an orphaned lock
				   file from an aborted previous run to block,
				   or result in error; rather, we only block
				   below, after the fork, when we know we may
				   have a legitimate lock file; note that
				   any orphaned lock files are removed after
				   each alarm completion; for debugging
				   purposes, we now write the exprstr to the
				   lock file
				 */
				if (mklock(execlockfile, exprstr, interactive, TRUE) == ERROR) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s(), %s, mklock() error for %s", fnnam, ALMNAM, execlockfile);
					/* no return: if we can't make the
					   lock file, we'll just treat this
					   like a simple EXEC (no wait)
					return TRUE;
					 */
				}
				switch (pid = fork()) {
					case -1:	/* error */
						logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in %s(), %s, fork() error", fnnam, ALMNAM);
						break;
					case 0:		/* child */
						cmd = (char *) mymalloc(strlen(exprstr)+strlen(piktlog)+strlen(RMCMD)+strlen(execlockfile)+24, fnnam);
						sprintf(cmd, "%s 2>> %s; %s -f %s 2>/dev/null", exprstr, piktlog, RMCMD, execlockfile);
						if (isautosyslog)
							logreturn(SYSLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "in %s(), %s, exec wait \"%s\"", fnnam, ALMNAM, cmd);
						logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "in %s(), %s, exec wait \"%s\"", fnnam, ALMNAM, cmd);
						/* exec the cmd, then
						   remove the lock
						 */
						execl(SHELL, "sh", "-c", cmd, (char *) NULL);
						free(cmd);
						_exit(127);
						break;
					default:	/* parent */
						/* wait until the child
						   removes the lock
						 */
						while (islock(execlockfile)) {
							sleep(EXECWAITWAIT);
						}
						break;
				}
				return TRUE;
			}
			break;
	}
	return QUIT;
}
